from __future__ import annotations

import asyncio
import csv
import io
import logging
import os
import re
from typing import TYPE_CHECKING, Tuple

import aiofiles
import rich

from cyberdrop_dl.base_functions.data_classes import MediaItem
from cyberdrop_dl.base_functions.error_classes import NoExtensionFailure

if TYPE_CHECKING:
    from pathlib import Path

    from yarl import URL
    
    from cyberdrop_dl.base_functions.sql_helper import SQLHelper


FILE_FORMATS = {
    'Images': {
        '.jpg', '.jpeg', '.png', '.gif',
        '.gifv', '.webp', '.jpe', '.svg',
        '.jfif', '.tif', '.tiff', '.jif',
    },
    'Videos': {
        '.mpeg', '.avchd', '.webm', '.mpv',
        '.swf', '.avi', '.m4p', '.wmv',
        '.mp2', '.m4v', '.qt', '.mpe',
        '.mp4', '.flv', '.mov', '.mpg',
        '.ogg', '.mkv', '.mts', '.ts',
        '.f4v'
    },
    'Audio': {
        '.mp3', '.flac', '.wav', '.m4a',
    }
}

logger = logging.getLogger(__name__)
MAX_FILENAME_LENGTH = 95


async def clear() -> None:
    """Clears the terminal screen"""
    clear_screen_proc = await asyncio.create_subprocess_shell('cls' if os.name == 'nt' else 'clear')
    await clear_screen_proc.wait()


def log(text: str, quiet: bool = False, style: str = "") -> None:
    """Logs to the output log file and optionally (by default) prints to the terminal with given style"""
    logger.debug(text)
    if not quiet:
        if style:
            text = f"[{style}]{text}[/{style}]"
        rich.print(text)


async def purge_dir(dirname: Path) -> None:
    """Purges empty directories"""
    deleted = []
    dir_tree = list(os.walk(dirname, topdown=False))

    for tree_element in dir_tree:
        sub_dir = tree_element[0]
        dir_count = len(os.listdir(sub_dir))
        if dir_count == 0:  # Helps with readability and i've had issues with it deleting non-empty dirs
            deleted.append(sub_dir)
    list(map(os.rmdir, deleted))


async def sanitize(name: str) -> str:
    """Simple sanitization to remove illegal characters"""
    return re.sub(r'[<>:"/\\|?*\']', "", name).strip()


async def make_title_safe(title: str) -> str:
    """Simple sanitization to remove illegal characters from titles and trim the length to be less than 60 chars"""
    title = title.replace("\n", "").strip()
    title = title.replace("\t", "").strip()
    title = re.sub(' +', ' ', title)
    title = re.sub(r'[\\*?:"<>|./]', "-", title)
    title = title[:60].strip()
    return title


async def check_direct(url: URL) -> bool:
    """Checks whether the given url is a direct link to a content item"""
    mapping_direct = [r'i.pixl.li', r'i..pixl.li', r'img-...cyberdrop...', r'f.cyberdrop...',
                      r'fs-...cyberdrop...', r'jpg.church/images/...', r'simp..jpg.church', r'jpg.fish/images/...',
                      r'simp..jpg.fish', r'jpg.fishing/images/...', r'simp..jpg.fishing', r's..putmega.com',
                      r's..putme.ga', r'images..imgbox.com', r's..lovefap...', r'img.kiwi/images/',
                      r'thumbs.gfycat.com/', r'giant.gfycat.com/']
    return any(re.search(domain, str(url)) for domain in mapping_direct)


async def get_filename_and_ext(filename: str, forum: bool = False) -> Tuple[str, str]:
    """Returns the filename and extension of a given file, throws NoExtensionFailure if there is no extension"""
    filename_parts = filename.rsplit('.', 1)
    if len(filename_parts) == 1:
        raise NoExtensionFailure()
    if filename_parts[-1].isnumeric() and forum:
        filename_parts = filename_parts[0].rsplit('-', 1)
    ext = "." + filename_parts[-1].lower()
    filename = filename_parts[0][:MAX_FILENAME_LENGTH] if len(filename_parts[0]) > MAX_FILENAME_LENGTH else filename_parts[0]
    filename = filename.strip()
    filename = await sanitize(filename + ext)
    return filename, ext


async def create_media_item(url: URL, referer: URL, sql_helper: SQLHelper, domain: str) -> MediaItem:
    """Returns the MediaItem of a given url, throws NoExtensionFailure if url.name doesn't have extension"""
    filename, ext = await get_filename_and_ext(url.name)
    complete = await sql_helper.check_complete_singular(domain, url)
    return MediaItem(url, referer, complete, filename, ext, filename)


class ErrorFileWriter:
    """Writes errors to a file"""
    def __init__(self, output_errored: bool, output_unsupported: bool, output_last_post: bool, errored_scrapes: Path,
                 errored_downloads: Path, unsupported: Path, last_post: Path):
        self.output_errored = output_errored
        self.output_unsupported = output_unsupported
        self.output_last_post = output_last_post

        self.errored_scrapes = errored_scrapes
        self.errored_downloads = errored_downloads
        self.unsupported = unsupported
        self.last_post = last_post

    async def row_to_csv(self, row: list[str]) -> str:
        output = io.StringIO()
        writer = csv.writer(output, quoting=csv.QUOTE_MINIMAL, lineterminator="\n")
        writer.writerow(row)
        return output.getvalue()

    async def write_errored_scrape(self, url: URL, e: Exception, quiet: bool) -> None:
        """Writes to the error file"""
        log(f"Error: {url}", quiet=quiet, style="red")
        logger.debug(e)

        if not self.output_errored:
            return

        async with aiofiles.open(self.errored_scrapes, 'a') as f:
            line = await self.row_to_csv([f"{url}", f"{e}"])
            await f.write(line)

    async def write_errored_scrape_header(self):
        """Writes to the error file"""
        if not self.output_errored:
            return

        async with aiofiles.open(self.errored_scrapes, 'a') as f:
            await f.write("URL,Exception\n")

    async def write_errored_download(self, url: URL, referer: URL, error_message: str) -> None:
        """Writes to the error file"""
        if not self.output_errored:
            return

        async with aiofiles.open(self.errored_downloads, 'a') as f:
            line = await self.row_to_csv([f"{url}", f"{referer}", error_message])
            await f.write(line)

    async def write_errored_download_header(self):
        """Writes to the error file"""
        if not self.output_errored:
            return

        async with aiofiles.open(self.errored_downloads, 'a') as f:
            await f.write("URL,REFERER,REASON\n")

    async def write_unsupported(self, url: URL, referer: URL, title: str) -> None:
        """Writes to the error file"""
        if not self.output_unsupported:
            return

        async with aiofiles.open(self.unsupported, 'a') as f:
            line = await self.row_to_csv([f"{url}", f"{referer}", title])
            await f.write(line)

    async def write_unsupported_header(self):
        """Writes to the error file"""
        if not self.output_unsupported:
            return

        async with aiofiles.open(self.unsupported, 'a') as f:
            await f.write("URL,REFERER,TITLE\n")

    async def write_last_post(self, url: URL) -> None:
        """Writes to the error file"""
        if not self.output_last_post:
            return

        async with aiofiles.open(self.last_post, 'a') as f:
            line = await self.row_to_csv([f"{url}"])
            await f.write(line)



from __future__ import annotations

import contextlib
from typing import Dict, List, Optional, Tuple

from rich.console import Group
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    SpinnerColumn,
    TaskID,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from rich.table import Table


def adjust_title(s: str, length: int = 40, placeholder: str = "...") -> str:
    """Collapse and truncate or pad the given string to fit in the given length"""
    return f"{s[:length - len(placeholder)]}{placeholder}" if len(s) >= length else s.ljust(length)


class Progressions:
    def __init__(self):
        self.overall_file_progress = Progress("[progress.description]{task.description}",
                                              BarColumn(bar_width=None),
                                              "[progress.percentage]{task.percentage:>3.2f}%",
                                              "{task.completed} of {task.total} Files Completed")

        self.forum_progress = Progress("[progress.description]{task.description}",
                                       BarColumn(bar_width=None),
                                       "[progress.percentage]{task.percentage:>3.2f}%",
                                       "{task.completed} of {task.total} Threads Completed")
        self.forum_progress_overflow = Progress("[progress.description]{task.description}")

        self.cascade_progress = Progress("[progress.description]{task.description}",
                                         BarColumn(bar_width=None),
                                         "[progress.percentage]{task.percentage:>3.2f}%",
                                         "{task.completed} of {task.total} Domains Completed")
        self.cascade_progress_overflow = Progress("[progress.description]{task.description}")

        self.domain_progress = Progress("[progress.description]{task.description}",
                                        BarColumn(bar_width=None),
                                        "[progress.percentage]{task.percentage:>3.2f}%",
                                        "{task.completed} of {task.total} Albums Completed")
        self.domain_progress_overflow = Progress("[progress.description]{task.description}")

        self.album_progress = Progress("[progress.description]{task.description}",
                                       BarColumn(bar_width=None),
                                       "[progress.percentage]{task.percentage:>3.2f}%",
                                       "{task.completed} of {task.total} Files Completed")
        self.album_progress_overflow = Progress("[progress.description]{task.description}")

        self.file_progress = Progress(SpinnerColumn(),
                                      "[progress.description]{task.description}",
                                      BarColumn(bar_width=None),
                                      "[progress.percentage]{task.percentage:>3.2f}%",
                                      "━",
                                      DownloadColumn(),
                                      "━",
                                      TransferSpeedColumn(),
                                      "━",
                                      TimeRemainingColumn())
        self.file_progress_overflow = Progress("[progress.description]{task.description}")

    async def get_overall_progress(self) -> Panel:
        return Panel(Group(self.overall_file_progress), title="Overall Progress", border_style="green", padding=(1, 1))

    async def get_forum_progress(self) -> Panel:
        return Panel(Group(self.forum_progress), title="Total Threads", border_style="green",padding=(1, 1))

    async def get_cascade_progress(self) -> Panel:
        return Panel(Group(self.cascade_progress, self.cascade_progress_overflow), title="Current Threads", border_style="green", padding=(1, 1))

    async def get_domain_progress(self) -> Panel:
        return Panel(Group(self.domain_progress, self.domain_progress_overflow), title="Current Domains", border_style="green", padding=(1, 1))

    async def get_album_progress(self) -> Panel:
        return Panel(Group(self.album_progress, self.album_progress_overflow), title="Current Albums", border_style="green", padding=(1, 1))

    async def get_file_progress(self) -> Panel:
        return Panel(Group(self.file_progress, self.file_progress_overflow), title="Current Downloads", border_style="green", padding=(1, 1))


class OverallFileProgress:
    """Class that keeps track of completed, skipped and failed files"""

    def __init__(self, total_files: int, overall_file_progress: Progress):
        self.overall_file_progress = overall_file_progress

        self.completed_files_task_id = overall_file_progress.add_task("[green]Completed", total=total_files)
        self.completed_files = 0
        self.skipped_files_task_id = overall_file_progress.add_task("[yellow]Skipped", total=total_files)
        self.skipped_files = 0
        self.failed_files_task_id = overall_file_progress.add_task("[red]Failed", total=total_files)
        self.failed_files = 0

    async def update_total(self, total_files: int) -> None:
        self.overall_file_progress.update(self.completed_files_task_id, total=total_files)
        self.overall_file_progress.update(self.skipped_files_task_id, total=total_files)
        self.overall_file_progress.update(self.failed_files_task_id, total=total_files)

    async def add_completed(self) -> None:
        self.overall_file_progress.advance(self.completed_files_task_id, 1)
        self.completed_files += 1

    async def add_skipped(self) -> None:
        self.overall_file_progress.advance(self.skipped_files_task_id, 1)
        self.skipped_files += 1

    async def add_failed(self) -> None:
        self.overall_file_progress.advance(self.failed_files_task_id, 1)
        self.failed_files += 1

    async def return_totals(self) -> Tuple[int, int, int]:
        return self.completed_files, self.skipped_files, self.failed_files

    async def hide(self) -> None:
        self.overall_file_progress.update(self.completed_files_task_id, visible=False)
        self.overall_file_progress.update(self.skipped_files_task_id, visible=False)
        self.overall_file_progress.update(self.failed_files_task_id, visible=False)


class _Progress:
    def __init__(self, progress: Progress, overflow: Progress, color: str, type_str: str, visible_tasks_limit: int):
        self.color = color
        self.type_str = type_str

        self.progress = progress
        self.progress_str = "[{color}]{description}"
        self.overflow = overflow
        self.overflow_str = "[{color}]... And {number} Other {type_str}"
        self.overflow_task_id = self.overflow.add_task(self.overflow_str.format(color=self.color, number=0, type_str=self.type_str), visible=False)

        self.locked = False

        self.visible_tasks: List[TaskID] = []
        self.invisible_tasks: List[TaskID] = []
        self.completed_tasks: List[TaskID] = []
        self.uninitiated_tasks: List[TaskID] = []
        self.tasks_visibility_limit = visible_tasks_limit

    async def redraw(self):
        while len(self.visible_tasks) > self.tasks_visibility_limit:
            task_id = self.visible_tasks.pop(0)
            self.invisible_tasks.append(task_id)
            self.progress.update(task_id, visible=False)
        while len(self.invisible_tasks) > 0 and len(self.visible_tasks) < self.tasks_visibility_limit:
            task_id = self.invisible_tasks.pop(0)
            self.visible_tasks.append(task_id)
            self.progress.update(task_id, visible=True)
        if len(self.invisible_tasks) > 0:
            self.overflow.update(self.overflow_task_id, description=self.overflow_str.format(color=self.color, number=len(self.invisible_tasks), type_str=self.type_str), visible=True)
        else:
            self.overflow.update(self.overflow_task_id, visible=False)

    async def add_task(self, description: str, total: int, initiated: bool) -> TaskID:
        if not initiated:
            task_id = self.progress.add_task(self.progress_str.format(color=self.color, description=description), total=total, visible=False)
            self.uninitiated_tasks.append(task_id)
        elif len(self.visible_tasks) >= self.tasks_visibility_limit:
            task_id = self.progress.add_task(self.progress_str.format(color=self.color, description=description), total=total, visible=False)
            self.invisible_tasks.append(task_id)
        else:
            task_id = self.progress.add_task(self.progress_str.format(color=self.color, description=description), total=total)
            self.visible_tasks.append(task_id)
        return task_id

    async def remove_task(self, task_id: TaskID) -> None:
        if task_id in self.visible_tasks:
            self.visible_tasks.remove(task_id)
            self.progress.update(task_id, visible=False)
        elif task_id in self.invisible_tasks:
            self.invisible_tasks.remove(task_id)
        elif task_id == self.overflow_task_id:
            self.overflow.update(task_id, visible=False)
        else:
            raise ValueError("Task ID not found")
        await self.redraw()

    async def mark_task_completed(self, task_id: TaskID) -> None:
        self.progress.update(task_id, visible=False)
        if task_id in self.visible_tasks:
            self.visible_tasks.remove(task_id)
        elif task_id in self.invisible_tasks:
            self.invisible_tasks.remove(task_id)
        await self.redraw()
        self.completed_tasks.append(task_id)

    async def advance_task(self, task_id: TaskID, amount: int) -> None:
        if task_id in self.uninitiated_tasks:
            self.uninitiated_tasks.remove(task_id)
            self.invisible_tasks.append(task_id)
            await self.redraw()
        self.progress.advance(task_id, amount)

    async def update_total(self, task_id: TaskID, total: int) -> None:
        if task_id in self.invisible_tasks:
            self.progress.update(task_id, total=total, visible=False)
        elif task_id in self.visible_tasks:
            self.progress.update(task_id, total=total, visible=True)
        elif task_id in self.uninitiated_tasks:
            self.progress.update(task_id, total=total, visible=False)
            self.uninitiated_tasks.remove(task_id)
            self.invisible_tasks.append(task_id)
            await self.redraw()


class ForumProgress:
    def __init__(self, progressions: Progressions):
        self.forum_progress = progressions.forum_progress
        self.color = "orange3"
        self.type_str = "Threads"
        self.progress = _Progress(self.forum_progress, progressions.forum_progress_overflow, self.color, "Threads", 5)

    async def add_forum(self, total_tasks: int) -> TaskID:
        return await self.progress.add_task("FORUM THREADS", total_tasks, True)

    async def advance_forum(self, task_id: TaskID) -> None:
        await self.progress.advance_task(task_id, 1)


class CascadeProgress:
    """Class that manages forum progression"""

    def __init__(self, progressions: Progressions, visible_tasks_limit: int):
        self.cascade_progress = progressions.cascade_progress
        self.color = "light_salmon3"
        self.type_str = "Threads"
        self.progress = _Progress(self.cascade_progress, progressions.cascade_progress_overflow, self.color, "Threads", visible_tasks_limit)

    async def add_cascade(self, title: str, total_domains: int) -> TaskID:
        task_id = await self.progress.add_task(title.upper(), total_domains, True)
        await self.progress.redraw()
        return task_id

    async def advance_cascade(self, task_id: TaskID) -> None:
        await self.progress.advance_task(task_id, 1)

    async def mark_cascade_completed(self, task_id: TaskID) -> None:
        await self.progress.mark_task_completed(task_id)
        await self.progress.redraw()


class DomainProgress:
    def __init__(self, progressions: Progressions, visible_tasks_limit: int):
        self.domain_progress = progressions.domain_progress
        self.color = "light_pink3"
        self.type_str = "Domains"
        self.progress = _Progress(self.domain_progress, progressions.domain_progress_overflow, self.color, "Domains", visible_tasks_limit)

        self.domains: Dict[str, TaskID] = {}
        self.domain_totals: Dict[str, int] = {}

    async def add_domain(self, domain: str, total_albums: int) -> TaskID:
        if domain in self.domains:
            self.domain_totals[domain] += total_albums
            await self.progress.update_total(self.domains[domain], self.domain_totals[domain])
        else:
            self.domains[domain] = await self.progress.add_task(domain.upper(), total_albums, True)
            self.domain_totals[domain] = total_albums

        await self.progress.redraw()
        return self.domains[domain]

    async def advance_domain(self, task_id: TaskID) -> None:
        await self.progress.advance_task(task_id, 1)

    async def mark_domain_completed(self, domain: str, task_id: TaskID) -> None:
        task = [x for x in self.domain_progress.tasks if x.id == task_id][0]
        if task.finished:
            with contextlib.suppress(KeyError):
                self.domains.pop(domain)
                self.domain_totals.pop(domain)
            await self.progress.mark_task_completed(task_id)


class AlbumProgress:
    def __init__(self, progressions: Progressions, visible_tasks_limit: int):
        self.album_progress = progressions.album_progress
        self.color = "pink3"
        self.type_str = "Albums"
        self.progress = _Progress(self.album_progress, progressions.album_progress_overflow, self.color, "Albums", visible_tasks_limit)

        self.albums: Dict[str, TaskID] = {}
        self.albums_totals: Dict[str, int] = {}

    async def add_album(self, album: str, total_files: int) -> TaskID:
        task_description = album.split('/')[-1]
        task_description = task_description.encode("ascii", "ignore").decode().strip()
        task_description = adjust_title(task_description).upper().strip()

        if task_description in self.albums:
            self.albums_totals[task_description] += total_files
            await self.progress.update_total(self.albums[task_description], self.albums_totals[task_description])
        else:
            self.albums[task_description] = await self.progress.add_task(task_description, total_files, True)
            self.albums_totals[task_description] = total_files

        await self.progress.redraw()
        return self.albums[task_description]

    async def advance_album(self, task_id: TaskID) -> None:
        await self.progress.advance_task(task_id, 1)

    async def mark_album_completed(self, album: str, task_id: TaskID) -> None:
        task_description = album.split('/')[-1]
        task_description = task_description.encode("ascii", "ignore").decode().strip()
        task_description = adjust_title(task_description).upper()

        task = [x for x in self.album_progress.tasks if x.id == task_id][0]
        if task.finished:
            with contextlib.suppress(KeyError):
                self.albums.pop(task_description)
                self.albums_totals.pop(task_description)
            await self.progress.mark_task_completed(task_id)


class FileProgress:
    def __init__(self, progressions: Progressions, visible_tasks_limit: int):
        self.file_progress = progressions.file_progress
        self.color = "plum3"
        self.type_str = "Files"
        self.progress = _Progress(self.file_progress, progressions.file_progress_overflow, self.color, "Files", visible_tasks_limit)

    async def add_file(self, file: str, expected_size: Optional[int]) -> TaskID:
        task_description = file.split('/')[-1]
        task_description = task_description.encode("ascii", "ignore").decode().strip()
        task_description = adjust_title(task_description)

        task_id = await self.progress.add_task(task_description.upper(), expected_size if expected_size else 0, False)
        await self.progress.redraw()
        return task_id

    async def update_file_length(self, task_id: TaskID, length: int) -> None:
        await self.progress.update_total(task_id, length)

    async def advance_file(self, task_id: TaskID, increment: int) -> None:
        await self.progress.advance_task(task_id, increment)

    async def remove_file(self, task_id: TaskID) -> None:
        await self.progress.remove_task(task_id)
        await self.progress.redraw()

    async def mark_file_completed(self, task_id: TaskID) -> None:
        await self.progress.mark_task_completed(task_id)
        await self.progress.redraw()


class ProgressMaster:
    def __init__(self, progress_options: Dict):
        self.hide_overall_progress = progress_options['hide_overall_progress']
        self.hide_forum_progress = progress_options['hide_forum_progress']
        self.hide_thread_progress = progress_options['hide_thread_progress']
        self.hide_domain_progress = progress_options['hide_domain_progress']
        self.hide_album_progress = progress_options['hide_album_progress']
        self.hide_file_progress = progress_options['hide_file_progress']
        self.refresh_rate = progress_options['refresh_rate']

        self.Progressions = Progressions()
        self.OverallFileProgress = OverallFileProgress(0, self.Progressions.overall_file_progress)
        self.ForumProgress = ForumProgress(self.Progressions)
        self.CascadeProgress = CascadeProgress(self.Progressions, progress_options['visible_rows_threads'])
        self.DomainProgress = DomainProgress(self.Progressions, progress_options['visible_rows_domains'])
        self.AlbumProgress = AlbumProgress(self.Progressions, progress_options['visible_rows_albums'])
        self.FileProgress = FileProgress(self.Progressions, progress_options['visible_rows_files'])

    async def get_table(self) -> Table:
        """Table creator for forum threads"""
        progress_table = Table.grid(expand=True)
        if not self.hide_overall_progress:
            progress_table.add_row(await self.Progressions.get_overall_progress())
        if not self.hide_forum_progress:
            progress_table.add_row(await self.Progressions.get_forum_progress())
        if not self.hide_thread_progress:
            progress_table.add_row(await self.Progressions.get_cascade_progress())
        if not self.hide_domain_progress:
            progress_table.add_row(await self.Progressions.get_domain_progress())
        if not self.hide_album_progress:
            progress_table.add_row(await self.Progressions.get_album_progress())
        if not self.hide_file_progress:
            progress_table.add_row(await self.Progressions.get_file_progress())
        return progress_table



from __future__ import annotations

import asyncio
import contextlib
import itertools
import logging
from http import HTTPStatus
from random import gauss
from typing import TYPE_CHECKING, Any, Dict, Tuple, Coroutine, List, Optional

import aiohttp.client_exceptions
from rich.live import Live

from cyberdrop_dl.base_functions.base_functions import (
    FILE_FORMATS,
    clear,
    log,
    logger,
)
from cyberdrop_dl.base_functions.data_classes import (
    AlbumItem,
    CascadeItem,
    DomainItem,
    FileLock,
    ForumItem,
    MediaItem,
)
from cyberdrop_dl.base_functions.error_classes import DownloadFailure
from cyberdrop_dl.base_functions.sql_helper import SQLHelper, get_db_path
from cyberdrop_dl.client.client import Client, DownloadSession

from .downloader_utils import (
    CustomHTTPStatus,
    allowed_filetype,
    basic_auth,
    check_free_space,
    get_threads_number,
    is_4xx_client_error,
    retry,
)
from .progress_definitions import (
    OverallFileProgress,
    ProgressMaster,
)

if TYPE_CHECKING:
    from pathlib import Path
    from rich.progress import TaskID

    from cyberdrop_dl.base_functions.base_functions import ErrorFileWriter


def _limit_concurrency(coroutines: List[Coroutine], semaphore: Optional[asyncio.Semaphore]) -> List[Coroutine]:
    if not semaphore:
        return coroutines

    async def limit_concurrency(coroutine: Coroutine) -> Coroutine:
        async with semaphore:
            return await coroutine

    return [limit_concurrency(coroutine) for coroutine in coroutines]


class CDLHelper:
    def __init__(self, args: Dict, client: Client, files: OverallFileProgress, SQL_Helper: SQLHelper,
                 error_writer: ErrorFileWriter):
        self.args = args

        # CDL Objects
        self.client = client
        self.files = files
        self.SQL_Helper = SQL_Helper
        self.error_writer = error_writer
        self.File_Lock = FileLock()

        # Limits
        self.delay = {'cyberdrop': 1.0, 'cyberfile': 1.0, 'anonfiles': 1.0, "coomer": 0.2, "kemono": 0.2}

        # Exclude Args
        self.exclude_audio = args["Ignore"]["exclude_audio"]
        self.exclude_images = args["Ignore"]["exclude_images"]
        self.exclude_videos = args["Ignore"]["exclude_videos"]
        self.exclude_other = args["Ignore"]["exclude_other"]

        # Runtime Args
        self.block_sub_folders = args['Runtime']['block_sub_folders']
        self.allowed_attempts = args["Runtime"]["attempts"]
        self.disable_attempt_limit = args["Runtime"]["disable_attempt_limit"]
        self.download_dir = args["Files"]["output_folder"]
        self.mark_downloaded = args["Runtime"]["skip_download_mark_completed"]
        self.proxy = args["Runtime"]["proxy"]
        self.required_free_space = args["Runtime"]["required_free_space"]

        # Filesize limits
        self.filesize_minimum_images = args["Runtime"]["filesize_minimum_images"]
        self.filesize_minimum_other = args["Runtime"]["filesize_minimum_other"]
        self.filesize_minimum_videos = args["Runtime"]["filesize_minimum_videos"]
        self.filesize_maximum_images = args["Runtime"]["filesize_maximum_images"]
        self.filesize_maximum_other = args["Runtime"]["filesize_maximum_other"]
        self.filesize_maximum_videos = args["Runtime"]["filesize_maximum_videos"]

        # Concurrency Limits
        self.threads_limit = asyncio.Semaphore(args["Runtime"]["max_concurrent_threads"]) if args["Runtime"][
            "max_concurrent_threads"] else None
        self.domains_limit = asyncio.Semaphore(args["Runtime"]["max_concurrent_domains"]) if args["Runtime"][
            "max_concurrent_domains"] else None
        self.albums_limit = asyncio.Semaphore(args["Runtime"]["max_concurrent_albums"]) if args["Runtime"][
            "max_concurrent_albums"] else None

        # API Keys
        self.pixeldrain_api_key = args["Authentication"]["pixeldrain_api_key"]

    def get_throttle(self, domain: str) -> float:
        """Get the throttle for a domain"""
        return self.delay.get(domain, self.client.throttle)

    async def check_filesize_limits(self, media: MediaItem, content_size: int) -> bool:
        if media.ext in FILE_FORMATS['Images']:
            if self.filesize_minimum_images and self.filesize_maximum_images:
                if content_size < self.filesize_minimum_images or content_size > self.filesize_maximum_images:
                    return False
            if content_size < self.filesize_minimum_images:
                return False
            if self.filesize_maximum_images and content_size > self.filesize_maximum_images:
                return False
        elif media.ext in FILE_FORMATS['Videos']:
            if self.filesize_minimum_videos and self.filesize_maximum_videos:
                if content_size < self.filesize_minimum_videos or content_size > self.filesize_maximum_videos:
                    return False
            if content_size < self.filesize_minimum_videos:
                return False
            if self.filesize_maximum_videos and content_size > self.filesize_maximum_videos:
                return False
        else:
            if self.filesize_minimum_other and self.filesize_maximum_other:
                if content_size < self.filesize_minimum_other or content_size > self.filesize_maximum_other:
                    return False
            if content_size < self.filesize_minimum_other:
                return False
            if self.filesize_maximum_other and content_size > self.filesize_maximum_other:
                return False
        return True


class Downloader:
    """Downloader class, directs downloading for domain objects"""

    def __init__(self, domain: str, CDL_Helper: CDLHelper, Progress_Master: ProgressMaster):
        self.domain = domain
        self.throttle = CDL_Helper.get_throttle(domain)

        max_workers = get_threads_number(CDL_Helper.args, domain)
        self._semaphore = asyncio.Semaphore(max_workers)
        self.current_attempt: Dict[str, int] = {}

        self.download_session = DownloadSession(CDL_Helper.client)

        self.CDL_Helper = CDL_Helper
        self.Progress_Master = Progress_Master

    async def download(self, album: str, media: MediaItem, url_path: str, album_task: TaskID) -> None:
        try:
            async with self._semaphore:
                await self.download_file(album, media, url_path, album_task)
        except RuntimeError as e:
            return

    @retry
    async def download_file(self, album: str, media: MediaItem, url_path: str, album_task: TaskID) -> None:
        """File downloader"""
        if not await check_free_space(self.CDL_Helper.required_free_space, self.CDL_Helper.download_dir):
            log("We've run out of free space.", quiet=True)
            await self.CDL_Helper.files.add_skipped()
            return

        if not await allowed_filetype(media, self.CDL_Helper.exclude_images, self.CDL_Helper.exclude_videos,
                                      self.CDL_Helper.exclude_audio, self.CDL_Helper.exclude_other):
            log(f"Blocked by file extension: {media.filename}", quiet=True)
            await self.CDL_Helper.files.add_skipped()
            return

        if self.CDL_Helper.block_sub_folders:
            album = album.split('/')[0]

        original_filename = media.original_filename
        filename = media.filename

        try:
            while await self.CDL_Helper.File_Lock.check_lock(filename):
                await asyncio.sleep(gauss(1, 1.5))
            await self.CDL_Helper.File_Lock.add_lock(filename)

            if url_path not in self.current_attempt:
                self.current_attempt[url_path] = 0

            complete_file = (self.CDL_Helper.download_dir / album / media.filename)
            partial_file = complete_file.with_suffix(complete_file.suffix + '.part')

            complete_file, partial_file, proceed, expected_size = await self.check_file_exists(complete_file,
                                                                                               partial_file, media,
                                                                                               album, url_path,
                                                                                               original_filename,
                                                                                               self.throttle)

            fake_download = self.CDL_Helper.mark_downloaded or not proceed

            await self.CDL_Helper.SQL_Helper.update_pre_download(complete_file, media.filename, url_path,
                                                                 original_filename)

            filesize_check = await self.CDL_Helper.check_filesize_limits(media, expected_size)
            if not filesize_check:
                log(f"Filesize out of specified range: {media.url}", quiet=True)
                await self.CDL_Helper.files.add_skipped()
                await self.CDL_Helper.File_Lock.remove_lock(filename)
                return

            await self.CDL_Helper.SQL_Helper.sql_insert_temp(str(partial_file))
            resume_point = partial_file.stat().st_size if partial_file.exists() else 0

            assert media.url.host is not None
            headers = {}
            if self.CDL_Helper.pixeldrain_api_key and "pixeldrain" in media.url.host:
                headers["Authorization"] = await basic_auth("Cyberdrop-DL", self.CDL_Helper.pixeldrain_api_key)

            headers['Range'] = f'bytes={resume_point}-'

            file_task = await self.Progress_Master.FileProgress.add_file(media.filename, expected_size)

            if not await self.CDL_Helper.SQL_Helper.sql_check_old_existing(url_path) and not fake_download:
                await self.download_session.download_file(self.Progress_Master, media, partial_file, self.throttle,
                                                          resume_point, self.CDL_Helper.proxy, headers, file_task)
                partial_file.rename(complete_file)

            await self.CDL_Helper.SQL_Helper.mark_complete(url_path, original_filename)
            if media.url.parts[-1] in self.current_attempt:
                self.current_attempt.pop(media.url.parts[-1])

            if fake_download:
                log(f"Already Downloaded: {media.filename} from {media.referer}", quiet=True)
                await self.CDL_Helper.files.add_skipped()
            else:
                await self.CDL_Helper.files.add_completed()
            await self.Progress_Master.FileProgress.mark_file_completed(file_task)

            log(f"Completed Download: {media.filename} from {media.referer}", quiet=True)
            await self.CDL_Helper.File_Lock.remove_lock(filename)
            return

        except (aiohttp.client_exceptions.ClientPayloadError, aiohttp.client_exceptions.ClientOSError,
                aiohttp.client_exceptions.ServerDisconnectedError, asyncio.TimeoutError,
                aiohttp.client_exceptions.ClientResponseError, aiohttp.client_exceptions.ServerTimeoutError,
                DownloadFailure, FileNotFoundError, PermissionError) as e:
            if await self.CDL_Helper.File_Lock.check_lock(filename):
                await self.CDL_Helper.File_Lock.remove_lock(filename)

            with contextlib.suppress(Exception):
                await self.Progress_Master.FileProgress.remove_file(file_task)

            if hasattr(e, "message"):
                logging.debug(f"\n{media.url} ({e.message})")

            if hasattr(e, "code"):
                if await is_4xx_client_error(e.code) and e.code != HTTPStatus.TOO_MANY_REQUESTS:
                    logger.debug("We ran into a 400 level error: %s", e.code)
                    log(f"Failed Download: {media.filename}", quiet=True)
                    if url_path in self.current_attempt:
                        self.current_attempt.pop(url_path)
                    await self.handle_failed(media, e)
                    return

                if e.code == HTTPStatus.SERVICE_UNAVAILABLE or e.code == HTTPStatus.BAD_GATEWAY \
                        or e.code == CustomHTTPStatus.WEB_SERVER_IS_DOWN:
                    if hasattr(e, "message"):
                        if not e.message:
                            e.message = "Web server is down"
                        logging.debug(f"\n{media.url} ({e.message})")
                    log(f"Failed Download: {media.filename}", quiet=True)
                    if url_path in self.current_attempt:
                        self.current_attempt.pop(url_path)
                    await self.handle_failed(media, e)
                    return

                logger.debug("Error status code: %s", e.code)

            raise DownloadFailure(code=getattr(e, "code", 1), message=repr(e))

    def can_retry(self, url_path: str) -> bool:
        return self.CDL_Helper.disable_attempt_limit or self.current_attempt[
            url_path] < self.CDL_Helper.allowed_attempts - 1

    async def handle_failed(self, media: MediaItem, e: Any) -> None:
        await self.CDL_Helper.files.add_failed()
        await self.CDL_Helper.error_writer.write_errored_download(media.url, media.referer, e.message)

    async def check_file_exists(self, complete_file: Path, partial_file: Path, media: MediaItem, album: str,
                                url_path: str, original_filename: str,
                                current_throttle: float) -> tuple[Path, Path, bool, int]:
        """Complicated checker for if a file already exists, and was already downloaded"""
        expected_size = None
        proceed = True
        while True:
            if not expected_size:
                expected_size = await self.download_session.get_filesize(media.url, str(media.referer),
                                                                         current_throttle)
            if not complete_file.exists() and not partial_file.exists():
                break

            if complete_file.exists() and complete_file.stat().st_size == expected_size:
                proceed = False
                break

            downloaded_filename = await self.CDL_Helper.SQL_Helper.get_downloaded_filename(url_path, original_filename)
            if not downloaded_filename:
                complete_file, partial_file = await self.iterate_filename(complete_file, media, album)
                break

            if media.filename == downloaded_filename:
                if partial_file.exists():
                    if partial_file.stat().st_size == expected_size:
                        proceed = False
                        partial_file.rename(complete_file)
                elif complete_file.exists():
                    if complete_file.stat().st_size == expected_size:
                        proceed = False
                    else:
                        complete_file, partial_file = await self.iterate_filename(complete_file, media, album)
                break

            media.filename = downloaded_filename
            complete_file = (self.CDL_Helper.download_dir / album / media.filename)
            partial_file = complete_file.with_suffix(complete_file.suffix + '.part')

        return complete_file, partial_file, proceed, expected_size

    async def iterate_filename(self, complete_file: Path, media: MediaItem, album: str) -> Tuple[Path, Path]:
        for iterations in itertools.count(1):
            filename = f"{complete_file.stem} ({iterations}){media.ext}"
            temp_complete_file = (self.CDL_Helper.download_dir / album / filename)
            if not temp_complete_file.exists() and not await self.CDL_Helper.SQL_Helper.check_filename(filename):
                media.filename = filename
                complete_file = (self.CDL_Helper.download_dir / album / media.filename)
                partial_file = complete_file.with_suffix(complete_file.suffix + '.part')
                break
        return complete_file, partial_file


class DownloadManager:
    def __init__(self, CDL_Helper: CDLHelper, Progress_Master: ProgressMaster):
        self.downloaders: Dict[str, Downloader] = {}
        self.CDL_Helper = CDL_Helper
        self.Progress_Master = Progress_Master

    async def get_downloader(self, domain: str) -> Downloader:
        if domain not in self.downloaders:
            self.downloaders[domain] = Downloader(domain, self.CDL_Helper, self.Progress_Master)
        return self.downloaders[domain]


class DownloadDirector:
    def __init__(self, args: Dict, Forums: ForumItem, SQL_Helper: SQLHelper, client: Client,
                 error_writer: ErrorFileWriter):
        self.Progress_Master = ProgressMaster(args["Progress_Options"])
        self.CDL_Helper = CDLHelper(args, client, self.Progress_Master.OverallFileProgress, SQL_Helper, error_writer)
        self.Download_Manager = DownloadManager(self.CDL_Helper, self.Progress_Master)

        self.Forums = Forums

    async def start(self):
        await self.CDL_Helper.files.update_total(await self.Forums.get_total())

        progress_table = await self.Progress_Master.get_table()
        with Live(progress_table, refresh_per_second=self.Progress_Master.refresh_rate):
            forum_task = await self.Progress_Master.ForumProgress.add_forum(len(self.Forums.threads))
            cascade_tasks = []
            for title, Cascade in self.Forums.threads.items():
                cascade_tasks.append(self.start_cascade(forum_task, title, Cascade))
            await asyncio.gather(*_limit_concurrency(cascade_tasks, self.CDL_Helper.threads_limit))

        await clear()
        completed_files, skipped_files, failed_files = await self.Progress_Master.OverallFileProgress.return_totals()
        log(f"| [green]Files Complete: {completed_files}[/green] - [yellow]Files Skipped:  {skipped_files}[/yellow] - [red]Files Failed: {failed_files}[/red] |")

        for downloader in self.Download_Manager.downloaders.values():
            await downloader.download_session.exit_handler()

    async def start_cascade(self, forum_task: TaskID, title: str, Cascade: CascadeItem) -> None:
        cascade_task = await self.Progress_Master.CascadeProgress.add_cascade(title, len(Cascade.domains))
        domain_tasks = []
        for domain, domain_obj in Cascade.domains.items():
            domain_tasks.append(self.start_domain(cascade_task, title, domain, domain_obj))
        await asyncio.gather(*_limit_concurrency(domain_tasks, self.CDL_Helper.domains_limit))
        await self.Progress_Master.CascadeProgress.mark_cascade_completed(cascade_task)
        await self.Progress_Master.ForumProgress.advance_forum(forum_task)

    async def start_domain(self, cascade_task: TaskID, title: str, domain: str, domain_obj: DomainItem) -> None:
        """Handler for domains and the progress bars for it"""
        downloader = await self.Download_Manager.get_downloader(domain)

        domain_task = await self.Progress_Master.DomainProgress.add_domain(domain, len(domain_obj.albums))
        album_tasks = []
        for album, album_obj in domain_obj.albums.items():
            album_tasks.append(self.start_album(downloader, domain_task, domain, album, album_obj))
        await asyncio.gather(*_limit_concurrency(album_tasks, self.CDL_Helper.albums_limit))
        await self.Progress_Master.CascadeProgress.advance_cascade(cascade_task)
        await self.Progress_Master.DomainProgress.mark_domain_completed(domain, domain_task)

    async def start_album(self, downloader: Downloader, domain_task: TaskID, domain: str, album: str,
                          album_obj: AlbumItem) -> None:
        """Handler for albums and the progress bars for it"""
        if await album_obj.is_empty():
            return

        album_task = await self.Progress_Master.AlbumProgress.add_album(album, len(album_obj.media))
        download_tasks = []
        for media in album_obj.media:
            download_tasks.append(self.start_file(downloader, album_task, domain, album, media))
        await asyncio.gather(*download_tasks)
        await self.Progress_Master.DomainProgress.advance_domain(domain_task)
        await self.Progress_Master.AlbumProgress.mark_album_completed(album, album_task)

    async def start_file(self, downloader: Downloader, album_task: TaskID, domain: str, album: str,
                         media: MediaItem) -> None:
        """Handler for files and the progress bars for it"""
        if media.complete or await self.CDL_Helper.SQL_Helper.check_complete_singular(domain, media.url):
            log(f"Previously Downloaded: {media.filename}", quiet=True)
            await self.CDL_Helper.files.add_skipped()
            await self.Progress_Master.AlbumProgress.advance_album(album_task)
            return

        url_path = await get_db_path(media.url, domain)
        await downloader.download(album, media, url_path, album_task)
        await self.Progress_Master.AlbumProgress.advance_album(album_task)
